//
//  UnzipJSBind.cpp
//  BXHall-mobile
//
//  Created by HBC on 2020/12/8.
//

#include "UnzipJSBind.h"
#include "cocos/cocos2d.h"
#include "scripting/js-bindings/manual/jsb_conversions.hpp"
#include "cocos/scripting/js-bindings/manual/jsb_classtype.hpp"
#include "platform/CCFileUtils.h"
#include "unzip/unzip.h"

#define BUFFER_SIZE 8192
#define MAX_FILENAME 512

USING_NS_CC;

static se::Object *__jsb_UnzipJSBind_proto = nullptr;
static se::Class *__jsb_UnzipJSBind_class = nullptr;

class UnzipJSBind
{
private:
    std::thread *_unzipThread;
    std::mutex _unzipMutex;

    uLong _doneTask;
    uLong _totalTask;

    bool _isError;

public:
    UnzipJSBind()
    {
        _unzipThread = nullptr;
        _doneTask = 0;
        _totalTask = 0;
        _isError = false;
    }

    ~UnzipJSBind()
    {
        delete _unzipThread;
        _unzipThread = nullptr;
    }

    /**
     *  自己添加的解压本地zip方法
     */
    bool decompressLocalZip(const std::string &zip)
    {
        CCLOG("AssetsManagerEx : decompressLocalZip  zip = %s  ...........", zip.c_str());

        _doneTask = 0;
        _totalTask = 0;
        _isError = false;
        _unzipThread = new std::thread(&UnzipJSBind::decompress, this, zip);
        //        std::thread t(&UnzipJSBind::decompress, this, zip);
        _unzipThread->detach();

        CCLOG("!!!线程结束后的操作！！！");

        return true;
    }

    bool decompress(const std::string &zip)
    {

        _unzipMutex.lock();

        FileUtils *fileUtils = FileUtils::getInstance();
        std::string writeablePath = fileUtils->getWritablePath();
        size_t pos_0 = zip.find_last_of("/\\");
        if (pos_0 == std::string::npos)
        {
            _isError = true;
            _unzipMutex.unlock();
            return false;
        }
        std::string zipName = zip.substr(pos_0 + 1, strlen(zip.c_str()));
        // remove directory
        size_t pos_b = zipName.find_first_of("_\\");
        if (pos_b == std::string::npos)
        {
            _isError = true;
            _unzipMutex.unlock();
            return false;
        }
        std::string bundleName = zipName.substr(0, pos_b);
        std::string bundleFullPath = writeablePath + "gamecaches/" + bundleName;
        if (fileUtils->isDirectoryExist(bundleFullPath))
        {
            fileUtils->removeDirectory(bundleFullPath);
        }
        std::string xmlPath = writeablePath + zipName;
        std::string curFullPath = fileUtils->fullPathForFilename(zip.c_str());
        // Find root path for zip file
        size_t pos = xmlPath.find_last_of("/\\");
        if (pos == std::string::npos)
        {
            //                CCLOG("AssetsManagerEx : no root path specified for curFullPath file %s\n", curFullPath.c_str());
            _isError = true;
            _unzipMutex.unlock();
            return false;
        }
        const std::string rootPath = xmlPath.substr(0, pos + 1) + "gamecaches/";
        //            CCLOG("AssetsManagerEx : rootPath  = %s  ...........1111" , rootPath.c_str());
        //            CCLOG("AssetsManagerEx : getSuitableFOpen  zip = %s  ..........." , fileUtils->getSuitableFOpen(curFullPath).c_str());
        CCLOG("writeable path is: %s", writeablePath.c_str());
        // Open the zip file
        unzFile zipfile = unzOpen(FileUtils::getInstance()->getSuitableFOpen(curFullPath).c_str());
        if (!zipfile)
        {
            //                CCLOG("AssetsManagerEx : can not open downloaded zip file %s\n", curFullPath.c_str());
            _isError = true;
            _unzipMutex.unlock();
            return false;
        }
        // Get info about the zip file
        unz_global_info global_info;
        if (unzGetGlobalInfo(zipfile, &global_info) != UNZ_OK)
        {
            //                CCLOG("AssetsManagerEx : can not read file global info of %s\n", curFullPath.c_str());
            _isError = true;
            unzClose(zipfile);
            _unzipMutex.unlock();
            return false;
        }
        // Buffer to hold data read from the zip file
        char readBuffer[BUFFER_SIZE];
        // Loop to extract all files.
        uLong i;
        _totalTask = global_info.number_entry;
        //            uLong doneTask = 0;
        //            CCLOG("Total task number: %lu", totalTask);

        for (i = 0; i < global_info.number_entry; ++i)
        {
            // Get info about current file.
            unz_file_info fileInfo;
            char fileName[MAX_FILENAME];
            if (unzGetCurrentFileInfo(zipfile,
                                      &fileInfo,
                                      fileName,
                                      MAX_FILENAME,
                                      NULL,
                                      0,
                                      NULL,
                                      0) != UNZ_OK)
            {
                CCLOG("AssetsManagerEx : can not read compressed file info\n");
                _isError = true;
                unzClose(zipfile);
                _unzipMutex.unlock();
                return false;
            }

            const std::string fullPath = rootPath + fileName;
            //                CCLOG("AssetsManagerEx : create directory fullPath %s\n", fullPath.c_str());

            // Check if this entry is a directory or a file.
            const size_t filenameLength = strlen(fileName);
            if (fileName[filenameLength - 1] == '/')
            {
                //There are not directory entry in some case.
                //So we need to create directory when decompressing file entry
                if (!fileUtils->createDirectory(basename(fullPath)))
                {
                    // Failed to create directory
                    CCLOG("AssetsManagerEx : can not create directory %s\n", fullPath.c_str());
                    _isError = true;
                    unzClose(zipfile);
                    _unzipMutex.unlock();
                    return false;
                }
                _doneTask++;
            }
            else
            {
                //  jsb depress zip xuyuanteng add 20170324
                std::string dir = basename(fullPath);
                //                     CCLOG("AssetsManagerEx : basename dir %s\n", dir.c_str());
                if (!fileUtils->isDirectoryExist(dir))
                {
                    CCLOG("AssetsManagerEx : isDirectoryExist   no  no  no");
                    if (!fileUtils->createDirectory(dir))
                    {
                        // Failed to create directory
                        CCLOG("AssetsManagerEx : can not create directory %s\n", fullPath.c_str());
                        _isError = true;
                        unzClose(zipfile);
                        _unzipMutex.unlock();
                        return false;
                    }
                }

                // Entry is a file, so extract it.
                // Open current file.
                if (unzOpenCurrentFile(zipfile) != UNZ_OK)
                {
                    CCLOG("AssetsManagerEx : can not extract file %s\n", fileName);
                    _isError = true;
                    unzClose(zipfile);
                    _unzipMutex.unlock();
                    return false;
                }

                // Create a file to store current file.
                //                    CCLOG("AssetsManagerEx : fullPath  =  %s\n", fullPath.c_str());
                //                    CCLOG("AssetsManagerEx : fopen getSuitableFOpen fullPath =  %s\n", FileUtils::getInstance()->getSuitableFOpen(fullPath).c_str());
                FILE *out = fopen(FileUtils::getInstance()->getSuitableFOpen(fullPath).c_str(), "wb");
                if (!out)
                {
                    CCLOG("AssetsManagerEx : can not create decompress destination file %s\n", fullPath.c_str());
                    _isError = true;
                    unzCloseCurrentFile(zipfile);
                    unzClose(zipfile);
                    _unzipMutex.unlock();
                    return false;
                }

                // Write current file content to destinate file.
                int error = UNZ_OK;
                do
                {
                    error = unzReadCurrentFile(zipfile, readBuffer, BUFFER_SIZE);
                    if (error < 0)
                    {
                        CCLOG("AssetsManagerEx : can not read zip file %s, error code is %d\n", fileName, error);
                        _isError = true;
                        fclose(out);
                        unzCloseCurrentFile(zipfile);
                        unzClose(zipfile);
                        _unzipMutex.unlock();
                        return false;
                    }

                    if (error > 0)
                    {
                        fwrite(readBuffer, error, 1, out);
                    }
                } while (error > 0);

                fclose(out);

                _doneTask++;
            }

            unzCloseCurrentFile(zipfile);

            // Goto next entry listed in the zip file.
            if ((i + 1) < global_info.number_entry)
            {
                if (unzGoToNextFile(zipfile) != UNZ_OK)
                {
                    CCLOG("AssetsManagerEx : can not read next file for decompressing\n");
                    _isError = true;
                    unzClose(zipfile);
                    _unzipMutex.unlock();
                    return false;
                }
            }
        }

        unzClose(zipfile);

        _unzipMutex.unlock();

        CCLOG("unzip finished!!!!!!");

        return true;
    }

    std::string basename(const std::string &path)
    {
        size_t found = path.find_last_of("/\\");
        if (std::string::npos != found)
        {
            return path.substr(0, found);
        }
        else
        {
            return path;
        }
    }

    std::string getUnzipState()
    {
        std::string state = "unzipping";
        if (_totalTask > 0 && _doneTask == _totalTask)
        {
            state = "done";
        }
        if (_isError ) {
            // 解压异常
            state = "error";
        }
        return std::to_string(_doneTask) + "/" + std::to_string(_totalTask) + "|" + state;
    }
};

static bool js_UnzipJSBind_decompressLocalZip(se::State &s)
{
    const auto &args = s.args();
    int argc = (int)args.size();
    if (argc >= 1)
    {
        UnzipJSBind *cobj = (UnzipJSBind *)s.nativeThisObject();
        se::Value zipUrl = args[0];
        cobj->decompressLocalZip(zipUrl.toString());
    }
    return true;
}
SE_BIND_FUNC(js_UnzipJSBind_decompressLocalZip)

static bool js_UnzipJSBind_getUnzipState(se::State &s)
{
    UnzipJSBind *cobj = (UnzipJSBind *)s.nativeThisObject();
    SE_PRECONDITION2(cobj, false, "js_UnzipJSBind_getUnzipState : Invalid Native Object");
    const auto &args = s.args();
    size_t argc = args.size();
    CC_UNUSED bool ok = true;
    if (argc == 0)
    {
        std::string result = cobj->getUnzipState();
        ok &= std_string_to_seval(result, &s.rval());
        SE_PRECONDITION2(ok, false, "js_UnzipJSBind_getUnzipState : Error processing arguments");
        return true;
    }
    SE_REPORT_ERROR("wrong number of arguments: %d, was expecting %d", (int)argc, 0);
    return false;
}
SE_BIND_FUNC(js_UnzipJSBind_getUnzipState)

static bool js_UnzipJSBind_finalize(se::State &s)
{
    UnzipJSBind *cobj = (UnzipJSBind *)s.nativeThisObject();
    delete cobj;
    cobj = NULL;
    return true;
}
SE_BIND_FINALIZE_FUNC(js_UnzipJSBind_finalize)

static bool js_UnzipJSBind_constructor(se::State &s)
{
    UnzipJSBind *cobj = new UnzipJSBind();
    s.thisObject()->setPrivateData(cobj);
    return true;
}
SE_BIND_CTOR(js_UnzipJSBind_constructor, __jsb_UnzipJSBind_class, js_UnzipJSBind_finalize)

bool js_register_UnzipJSBind(se::Object *global)
{
    auto cls = se::Class::create("UnzipJSBind", global, nullptr, _SE(js_UnzipJSBind_constructor));

    // 定义函数
    cls->defineFunction("decompressLocalZip", _SE(js_UnzipJSBind_decompressLocalZip));
    cls->defineFunction("getUnzipState", _SE(js_UnzipJSBind_getUnzipState));

    cls->defineFinalizeFunction(_SE(js_UnzipJSBind_finalize));
    cls->install();

    // 保存注册的结果，便于其他地方使用，比如类继承
    __jsb_UnzipJSBind_proto = cls->getProto();
    __jsb_UnzipJSBind_class = cls;

    se::ScriptEngine::getInstance()->clearException();
    return true;
}
